'''
TODO:
-Comment the code nicely

Project purpose is to import a gds and export a blender file.

This project depends heavily on:

PHIDL (https://github.com/amccaugh/phidl)
gdspy (https://github.com/heitzmann/gdspy/)

It is not intended as an individualistic feat, but more as a synthesis of simple
methods to create nice pictures. 

Author:
Matthew van Niekerk
'''
import numpy as np

try:
    import bpy
except:
    print('You must run this file from the Blender build of python to ustilize interop with Blender.')

from .layer import Layer, LayerStack
import matplotlib.pyplot as plt
import phidl.geometry as pg
import math
import phidl
from phidl.device_layout import Device, DeviceReference, Port, Polygon
import gdspy
# when this is run from blenders python, need to import the module in this way
import sys
# sys.path.append('/Users/matthewvanniekerk/Documents/RIT/GIT/GDS-BLENDER-PY/gds-blender-py/')
import os
#test

import sys
import warnings

from progressbar import *
  
'''
Importer object takes a gds file as a constrcutoor

'''
class Importer(object):
    '''
    Parameters
    ----------
    filename (string) - the filename of the gds file
    layerstack (LayerStack) - Layer Stack for this gds
    '''
    def __init__(self, filename, layerstack, cellname = None):

        self.filename = filename
        self.cellname = cellname
        self.layerstack = layerstack
        self.gds = self._import_gds()
        self._etch_target_helper()
        self.vertices = self._extract_vertices()


    '''
    Paramters
    ---------
    filename (string) - the filename of the GDS

    Returns
    -------
    D (list) - list of Device objects

    Purpose
    -------
    Import any gds file and select the largest top cell as the top cell.
    '''
    def _import_gds(self):
        gdsii_lib = gdspy.GdsLibrary()
        gdsii_lib.read_gds(self.filename)
        top_level_cells = gdsii_lib.top_level()
        cellname = self.cellname 
        if cellname is not None:
            if cellname not in gdsii_lib.cell_dict:
                raise ValueError('The requested cell (named %s) is not present in file %s' % (cellname,self.filename))
            topcell = gdsii_lib.cell_dict[cellname]
        elif cellname is None and len(top_level_cells) == 1:
            topcell = top_level_cells[0]
        elif cellname is None and len(top_level_cells) > 1:
            areas = []
            for cell in top_level_cells:
                areas.append(cell.area())
            ind = areas.index(max(areas))
            topcell = top_level_cells[ind]
            
        D = Device('import_gds')
        polygons = topcell.get_polygons(by_spec = True)

        for layer_in_gds, polys in polygons.items():
            D.add_polygon(polys, layer = layer_in_gds)
        return D
    '''
    Remove the drawn etch layers 
    '''
    def _remove_etch_layers(self):
        
        for lay in self.layerstack.etch_layers:
            self.gds.remove_layers(layers = (lay.layer,lay.datatype))

    '''
    Take the etch layer, boolean difference the two layers, and add 
    that layer to the layer that was removed.
    '''
    def _etch_target_helper(self):

        etch_layers = []
        etch_targets = []
        layer_target = []
        for elay in self.layerstack.etch_layers:
            etch_layers.append(self.gds.get_polygons(by_spec = True, depth = None)[(elay.layer,elay.datatype)])
            layer_target.append((elay.layer,elay.datatype))
        for etar in self.layerstack.etch_targets:
            etch_targets.append(self.gds.get_polygons(by_spec = True, depth = None)[(etar.layer,etar.datatype)])
            
        # Do this with gdspy! dont go back to Device.

        etch_layers,etch_targets
        new_polygons = []
        for i in range(len(etch_targets)):
            p = gdspy.fast_boolean(operand1 = etch_targets[i], \
                                operand2 = etch_layers[i], \
                                operation = 'not', \
                                precision = 1e-9, \
                                max_points = 4000,\
                                layer = layer_target[i][0],\
                                datatype = layer_target[i][1] )
            new_polygons.append(p)
        self._remove_etch_layers()

        for i in range(len(new_polygons)):
            self.gds.add_polygon(new_polygons[i],layer = layer_target[i])

    def _extract_vertices(self):

        extracted = {}
        points = []
        layers = []
        gds = self.gds
        
        if gds is not list: gds = [gds]
        for item in gds:
            if isinstance(item, (Device, DeviceReference, gdspy.CellArray)):
                polygons_spec = item.get_polygons(by_spec = True, depth = None)
                for key in sorted(polygons_spec):
                    polygons = polygons_spec[key]
                    layers.append(key)
                    points.append(self._vert_polygons(polygons))
            elif isinstance(item, Polygon):
                polygons = item.polygons
                layers.append(key)
                points.append(self._vert_polygons(polygons))     
        extracted['vertices'] = points
        extracted['layers'] = layers
        return extracted

    def _vert_polygons(self, polygons):
        vertices = []
        for poly in polygons:
            points = np.concatenate((poly,[poly[0]]),axis=0)
            points_stack = np.vstack(points)
            points_out = points_stack[:,0],points_stack[:,1]
            vertices.append(points_out)
        return vertices
    
    def draw_in_blender(self):
        # Delete all meshes in the scene
        for item in bpy.data.meshes:
            bpy.data.meshes.remove(item)
        vertices = self.vertices
        layer_stack = self.layerstack
        for layer in layer_stack.layers:
            mat = bpy.data.materials.get(layer.name)
            if mat is None:
                mat = bpy.data.materials.new(name=layer.name)
            mat.diffuse_color = layer.color
            
        vert_list = []
        extrude = []
        lname = []
        for i in range(len(vertices['vertices'])):
            vert = vertices['vertices'][i]
            
            z, thickness, layer_name = layer_stack.fetch_params(vertices['layers'][i])

            if z is None:
                z = -1
            for v in vert:
                vert_list.append(np.array((v[0],v[1],[z]*len(v[0]))).transpose())
                extrude.append(thickness)
                lname.append(layer_name)
                
        widgets = ['Drawing:', Percentage(), ' ', Bar(marker='#', left = '[', right = ']' ), \
                   ' ', ETA(), ' ', FileTransferSpeed()]
        pbar = ProgressBar(widgets = widgets, max_value= len(vert_list) + len(lname))
        pbar.start()
        for i in range(len(vert_list)):
            if lname[i] is not None:
                verts_data = vert_list[i]
                num_verts = len(verts_data)
                faces_data = [ np.linspace(0,num_verts-1,num_verts).astype(int) ]

                mesh = bpy.data.meshes.new('myMesh_mesh')
                mesh.from_pydata(verts_data,[],faces_data)
                mesh.update()

                new_object = bpy.data.objects.new(lname[i],mesh)
                new_object.data = mesh

                collection = bpy.context.collection
                collection.objects.link(new_object)
                bpy.context.view_layer.objects.active = new_object
                bpy.ops.object.editmode_toggle()

                bpy.ops.mesh.extrude_region_move(
                    TRANSFORM_OT_translate={"value":(0, 0, extrude[i] )}
                )
                bpy.ops.object.editmode_toggle()
        
                new_object.data.materials.append(bpy.data.materials.get(lname[i]))
                pbar.update(i)
                i = i + 1
        
        
        # Join All like materials to one Mesh
        context = bpy.context
        scene = context.scene
        for layer in lname:
            mats = [m.name for m in bpy.data.materials if m.name== layer]
            for mat in mats:
                obs = [o for o in scene.objects
                        if o.type == 'MESH'
                        and mat in o.material_slots]

                if len(obs) > 1:
                    # clear prior selection
                    for o in context.selected_objects:
                        o.select_set(False)
                    for o in obs:
                        o.select_set(True)
                    context.view_layer.objects.active = obs[0]
                    bpy.ops.object.join()
            i = i + 1
        pbar.finish()
        
        





        


from setuptools import setup,find_packages

with open('README.md', 'r') as f:
    readme = f.read()
dependencies = [
        'numpy',
        'gdspy',
        'matplotlib>=2.2.2',
        'phidl',
        'progressbar2'
]
setup(
    name='gdsblenderpy',
    version='0.0.2',
    description='GDS to Blender Python Translator',
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Matthew van Niekerk',
    author_email='matthewvanniekerk@mac.com',
    url='https://github.com/matthewvanniekerk/gds-blender-py',
    packages=find_packages(),
    install_requires=dependencies,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)


#from .importer import Importer
import matplotlib.pyplot as plt
import numpy as np

class Layer(object):
    '''
    Parameters
    -----------
    layer (int) - the layer number
    datatype (int) - the datatype
    z (float) - the starting z height in microns
    thickness (float) - layer thickness extending from z to final height
    color (float) - (r , g, b, alpha)
    '''
    def __init__(self, name, layer, datatype, z, thickness, color, alpha, etch_target = None ):
        
        self.name = name
        self.layer = layer
        self.datatype = datatype
        self.z = z
        self.thickness = thickness
        self.etch_target = etch_target
        self.color = (color[0]/255,color[1]/255,color[2]/255,alpha)
        
class LayerStack(object):
    '''
    Parameters
    ----------

    layers (list) - list of layers 
    '''
    def __init__(self,name,layers):
        self.name = name
        self.layers = layers
        self.etch_targets = []
        self.etch_layers = []
        # create the etch layer and targets
        for lay in self.layers:
            if lay.etch_target is not None:
                self.etch_targets.append(lay.etch_target)
                self.etch_layers.append(lay)
        # chcange the heights
        for lay in self.etch_layers:
            final_height = lay.etch_target.thickness + lay.etch_target.z
            lay.thickness = final_height - lay.z
            lay.etch_target.thickness = lay.z-lay.etch_target.z
            lay.color = lay.etch_target.color

    def plot(self):

        for lay in self.layers:
            plt.fill([0 , 0 , 1, 1],[lay.z,lay.z+lay.thickness, lay.z+lay.thickness, lay.z], label = lay.name, alpha = 0.3,  color = (lay.color[0],lay.color[1],lay.color[2]), lw = 0)
                
        plt.title(self.name)
        plt.grid(False)
        plt.xlabel('')
        plt.tick_params(
            axis='x',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            bottom=False,      # ticks along the bottom edge are off
            top=False,         # ticks along the top edge are off
            labelbottom=False)
        plt.ylabel(r'Layer Z Profile ($\mu$m)')
        plt.legend()
        plt.show()

    def fetch_params(self, key):
        for lay in self.layers:
            if key[0] == lay.layer and key[1] == lay.datatype:
                return lay.z, lay.thickness, lay.name
        return 0,0, None


